"""Coordinate Reference System module.

This module contains functions related to coordinate reference system (CRS)
selection and manipulation.  A CRS can also be called a spatial reference
system (SRS) and the names can be used interchangeably.

In this module, SRS is used to represent the Harmony Message projection
information as it has a named SRS object that is retrieved from the Harmony Message's
Format object.

"""

from harmony_service_lib.message import SRS
from pyproj.crs import CRS as pyCRS

# pylint: disable-next=no-name-in-module
from rasterio.crs import CRS
from xarray import DataArray

from hybig.exceptions import HyBIGValueError

# These are the CRSs that GIBS will accept as input. When the user hasn't
# directly specified an output CRS, the code will attempt to choose the best
# one of these.
PREFERRED_CRS = {
    'north': 'EPSG:3413',  # WGS 84 / NSIDC Sea Ice Polar Stereographic North
    'south': 'EPSG:3031',  # WGS 84 / Antarctic Polar Stereographic
    'global': 'EPSG:4326',  # WGS 84 -- WGS84 - World Geodetic System 1984, used in GPS
}


def choose_target_crs(srs: SRS, data_array: DataArray) -> CRS:
    """Return the target CRS for the output image.

    If a harmony message defines a SRS, we use that as the target ouptut CRS.
    Otherwise, we try determine the best "GIBS compatible" CRS based on the
    input granule metadata.

    """
    if srs is not None:
        return choose_crs_from_srs(srs)
    return choose_crs_from_metadata(data_array)


def choose_crs_from_srs(srs: SRS):
    """Choose the best CRS based on the Harmony Request's SRS.

    create a CRS from the harmony srs information.
    prefer epsg to wkt
    prefer wkt to proj4

    Raise HyBIGValueError if the harmony SRS cannot be converted to a
    rasterio CRS for any reason.

    """
    try:
        if srs.epsg is not None and srs.epsg != '':
            return CRS.from_string(srs.epsg)
        if srs.wkt is not None and srs.wkt != '':
            return CRS.from_string(srs.wkt)
        return CRS.from_string(srs.proj4)
    except Exception as exception:
        raise HyBIGValueError(f'Bad input SRS: {str(exception)}') from exception


def is_preferred_crs(crs: CRS) -> bool:
    """Returns true if the input CRS is preferred by GIBS."""
    if crs.to_string() in PREFERRED_CRS.values():
        return True
    return False


def choose_crs_from_metadata(data_array: DataArray) -> CRS | None:
    """Determine the best CRS based on input metadata."""
    if is_preferred_crs(data_array.rio.crs):
        return data_array.rio.crs
    return choose_best_crs_from_metadata(data_array.rio.crs)


def choose_best_crs_from_metadata(crs: CRS) -> CRS:
    """Determine the best preferred CRS based on the input CRS.

    We are targeting GIBS which has three preferred CRSs a Northern Polar
    Stereo, Southern Polar Stero and Global Projection. Using the information
    from the input granule we want to determine which CRS the dataset will be
    best represented by.

    This routine transforms the input CRS into a dict and examines the CRS
    projection parameters to determine the best option for regridding.

    It is very naive, if the proj is lonlat it uses the preferred global
    projection.

    If the projection latitude of origin is above 80° N -> use the preferred
    northern projection

    If the projection latitude of origin is below -80° N -> use the preferred
    southern projection

    Otherwise it defaults to the global projection.

    """
    projection_params = pyCRS(crs).to_dict()

    if projection_params.get('proj', None) == 'longlat':
        return CRS.from_string(PREFERRED_CRS['global'])

    if projection_params.get('lat_0', 0.0) >= 80:
        return CRS.from_string(PREFERRED_CRS['north'])

    if projection_params.get('lat_0', 0.0) <= -80:
        return CRS.from_string(PREFERRED_CRS['south'])

    return CRS.from_string(PREFERRED_CRS['global'])



"""Module containing core functionality for browse image generation."""

import re
from itertools import zip_longest
from logging import Logger, getLogger
from pathlib import Path

import matplotlib
import numpy as np
import rasterio
from affine import dumpsw
from harmony_service_lib.message import Message as HarmonyMessage
from harmony_service_lib.message import Source as HarmonySource
from matplotlib.colors import Normalize
from numpy import ndarray, uint8
from osgeo_utils.auxiliary.color_palette import ColorPalette
from PIL import Image
from rasterio.io import DatasetReader
from rasterio.warp import Resampling, reproject
from rioxarray import open_rasterio
from xarray import DataArray

from hybig.browse_utility import get_harmony_message_from_params
from hybig.color_utility import (
    NODATA_IDX,
    NODATA_RGBA,
    OPAQUE,
    TRANSPARENT,
    ColorMap,
    all_black_color_map,
    colormap_from_colors,
    get_color_palette,
    greyscale_colormap,
    palette_from_remote_colortable,
)
from hybig.exceptions import HyBIGError
from hybig.sizes import (
    GridParams,
    create_tiled_output_parameters,
    get_target_grid_parameters,
)


def create_browse(
    source_tiff: str,
    params: dict = None,
    palette: str | ColorPalette | None = None,
    logger: Logger = None,
) -> list[tuple[Path, Path, Path]]:
    """Create browse imagery from an input geotiff.

    This is the exposed library function to allow users to create browse images
    from the hybig-py library. It parses the input params and constructs the
    correct Harmony input structure [Message.Format] to call the service's
    entry point create_browse_imagery.

    Output images are created and deposited into the input GeoTIFF's directory.

    Args:
        source_tiff: str, location of the input geotiff to process.

        params: [dict | None], A dictionary with the following keys:

            mime: [str], MIME type of the output image (default: 'image/png').
                  any string that contains 'jpeg' will return a jpeg image,
                  otherwise create a png.

            crs: [dict | None], Target image's Coordinate Reference System.
                 A dictionary with 'epsg', 'proj4' or 'wkt' key.

            scale_extent: [dict | None], Scale Extents for the image. This dictionary
                contains "x" and "y" keys each whose value which is a dictionary
                of "min", "max" values in the same units as the crs.
                e.g.: { "x": { "min": 0.5, "max": 125 },
                        "y": { "min": 52, "max": 75.22 } }

            scale_size: [dict | None], Scale sizes for the image.  The dictionary
                contains "x" and "y" keys with the horizontal and veritcal
                resolution in the same units as the crs.
                e.g.: { "x": 10, "y": 10 }

            height: [int | None], height of the output image in gridcells.

            width: [int | none], width of the output image in gridcells.

        palette: [str | ColorPalette | none], either a URL to a remote color palette
             that is fetched and loaded or a ColorPalette object used to color
             the output browse image. If not provided, a grayscale image is
             generated.

        logger: [Logger | None], a configured Logger object. If None a default
             logger will be used.

    Note:
      if supplied, scale_size, scale_extent, height and width must be
      internally consistent.  To define a valid output grid:
            * Specify scale_extent and 1 of:
              * height and width
              * scale_sizes (in the x and y horizontal spatial dimensions)
            * Specify all three of the above, but ensure values are consistent
              with one another, noting that:
              scale_size.x = (scale_extent.x.max - scale_extent.x.min) / width
              scale_size.y = (scale_extent.y.max - scale_extent.y.min) / height

    Returns:
        List of 3-element tuples. These are the file paths of:
        - The output browse image
        - Its associated ESRI world file (containing georeferencing information)
        - The auxiliary XML file (containing duplicative georeferencing information)


    Example Usage:
        results = create_browse(
            "/path/to/geotiff",
            {
                "mime": "image/png",
                "crs": {"epsg": "EPSG:4326"},
                "scale_extent": {
                    "x": {"min": -180, "max": 180},
                    "y": {"min": -90, "max": 90},
                },
                "scale_size": {"x": 10, "y": 10},
            },
            "https://remote-colortable",
            logger,
        )

    """
    harmony_message = get_harmony_message_from_params(params)

    if logger is None:
        logger = getLogger('hybig-py')

    if isinstance(palette, str):
        color_palette = palette_from_remote_colortable(palette)
    else:
        color_palette = palette

    return create_browse_imagery(
        harmony_message, source_tiff, HarmonySource({}), color_palette, logger
    )


def create_browse_imagery(
    message: HarmonyMessage,
    input_file_path: str,
    source: HarmonySource,
    item_color_palette: ColorPalette | None,
    logger: Logger,
) -> list[tuple[Path, Path, Path]]:
    """Create browse image from input geotiff.

    Take input browse image and return a 3-element tuple for the file paths of
    the output browse image, its associated ESRI world file and the auxilary
    xml file.

    """
    output_driver = image_driver(message.format.mime)
    out_image_file = output_image_file(Path(input_file_path), driver=output_driver)
    out_world_file = output_world_file(Path(input_file_path), driver=output_driver)

    try:
        with open_rasterio(
            input_file_path, mode='r', mask_and_scale=True
        ) as rio_in_array:
            in_dataset = rio_in_array.rio._manager.acquire()
            validate_file_type(in_dataset)
            validate_file_crs(rio_in_array)

            if rio_in_array.rio.count == 1:
                color_palette = get_color_palette(
                    in_dataset, source, item_color_palette
                )
                if output_driver == 'JPEG':
                    # For JPEG output, convert to RGB
                    # color_map will be None
                    raster, color_map = convert_singleband_to_rgb(
                        rio_in_array, color_palette
                    )
                else:
                    # For PNG output, use palettized approach
                    raster, color_map = convert_singleband_to_raster(
                        rio_in_array, color_palette
                    )
            elif rio_in_array.rio.count in (3, 4):
                raster = convert_mulitband_to_raster(rio_in_array)
                color_map = None
                if output_driver == 'JPEG':
                    raster = raster[0:3, :, :]
            else:
                raise HyBIGError(
                    f'incorrect number of bands for image: {rio_in_array.rio.count}'
                )

            grid_parameters = get_target_grid_parameters(message, rio_in_array)
            grid_parameter_list, tile_locators = create_tiled_output_parameters(
                grid_parameters
            )

            processed_files = []
            for grid_parameters, tile_location in zip_longest(
                grid_parameter_list, tile_locators
            ):
                tiled_out_image_file = get_tiled_filename(out_image_file, tile_location)
                tiled_out_world_file = get_tiled_filename(out_world_file, tile_location)
                tiled_out_aux_xml_file = get_aux_xml_filename(tiled_out_image_file)
                logger.info(f'out image file: {tiled_out_image_file}: {tile_location}')

                write_georaster_as_browse(
                    rio_in_array,
                    raster,
                    color_map,
                    grid_parameters,
                    logger=logger,
                    driver=output_driver,
                    out_file_name=tiled_out_image_file,
                    out_world_name=tiled_out_world_file,
                )
                processed_files.append(
                    (tiled_out_image_file, tiled_out_world_file, tiled_out_aux_xml_file)
                )

    except Exception as exception:
        raise HyBIGError(str(exception)) from exception

    return processed_files


def convert_mulitband_to_raster(data_array: DataArray) -> ndarray[uint8]:
    """Convert multiband to a raster image.

    Return a 4-band raster, where the alpha layer is presumed to be the missing
    data mask.

    Convert 3-band data into a 4-band raster by generating an alpha layer from
    any missing data in the RGB bands.

    """
    if data_array.rio.count not in [3, 4]:
        raise HyBIGError(
            f'Cannot create image from {data_array.rio.count} band image. '
            'Expecting 3 or 4 bands.'
        )

    bands = data_array.to_numpy()

    if data_array.rio.count == 4:
        return convert_to_uint8(bands, original_dtype(data_array))

    # Input NaNs in any of the RGB bands are made transparent.
    nan_mask = np.isnan(bands).any(axis=0)
    nan_alpha = np.where(nan_mask, TRANSPARENT, OPAQUE)

    raster = convert_to_uint8(bands, original_dtype(data_array))

    return np.concatenate((raster, nan_alpha[None, ...]), axis=0)


def convert_to_uint8(bands: ndarray, dtype: str | None) -> ndarray[uint8]:
    """Convert Banded data with NaNs (missing) into a uint8 data cube.

    Nearly all of the time this will simply pass through the data coercing it
    back into unsigned ints and setting the missing values to 0 that will be
    masked as transparent in the output png.

    There is a some small non-zero chance that the input RGB image was 16-bit
    and if any of the values exceed 255, we must normalize all of input data to
    the range 0-255.

    """

    if dtype != 'uint8' and np.nanmax(bands) > 255:
        norm = Normalize(vmin=np.nanmin(bands), vmax=np.nanmax(bands))
        scaled = np.around(norm(bands) * 255.0)
        raster = scaled.filled(0).astype('uint8')
    else:
        raster = np.nan_to_num(bands).astype('uint8')

    return raster


def original_dtype(data_array: DataArray) -> str | None:
    """Return the original input data's type.

    rastero_open retains the input dtype in the encoding dictionary and is used
    to understand what kind of casts are safe.

    """
    return data_array.encoding.get('dtype') or data_array.encoding.get('rasterio_dtype')


def convert_singleband_to_raster(
    data_array: DataArray,
    color_palette: ColorPalette | None = None,
) -> tuple[ndarray, ColorMap]:
    """Convert input dataset to a 1-band palettized image with colormap.

    Uses a palette if provided otherwise returns a greyscale image.
    """
    if color_palette is None:
        return scale_grey_1band(data_array)
    return scale_paletted_1band(data_array, color_palette)


def scale_grey_1band(data_array: DataArray) -> tuple[ndarray, ColorMap]:
    """Normalize input array and return scaled data with greyscale ColorMap."""
    band = data_array[0, :, :]
    norm = Normalize(vmin=np.nanmin(band), vmax=np.nanmax(band))

    # Scale input data from 0 to 254
    normalized_data = norm(band) * 254.0

    # Set any missing to missing
    normalized_data[np.isnan(normalized_data)] = NODATA_IDX

    grey_colormap = greyscale_colormap()
    raster_data = np.expand_dims(np.round(normalized_data).data, 0)
    return np.array(raster_data, dtype='uint8'), grey_colormap


def convert_singleband_to_rgb(
    data_array: DataArray,
    color_palette: ColorPalette | None = None,
) -> tuple[ndarray, None]:
    """Convert input 1-band dataset to RGB image for JPEG output.

    Uses a palette if provided, otherwise returns a greyscale RGB image.
    Returns a 3-band RGB array and None for colormap (since RGB doesn't need colormap).
    """
    if color_palette is None:
        return scale_grey_1band_to_rgb(data_array)
    return scale_paletted_1band_to_rgb(data_array, color_palette)


def scale_grey_1band_to_rgb(data_array: DataArray) -> tuple[ndarray, None]:
    """Normalize input array and return as 3-band RGB grayscale image."""
    band = data_array[0, :, :]
    norm = Normalize(vmin=np.nanmin(band), vmax=np.nanmax(band))

    # Scale input data from 0 to 254 (palettized data is 254-level quantized)
    normalized_data = norm(band) * 254.0

    # Set any missing to 0 (black), no transparency
    normalized_data[np.isnan(normalized_data)] = 0

    grey_data = np.round(normalized_data).astype('uint8')
    rgb_data = np.stack([grey_data, grey_data, grey_data], axis=0)

    return rgb_data, None


def scale_paletted_1band_to_rgb(
    data_array: DataArray, palette: ColorPalette
) -> tuple[ndarray, None]:
    """Scale a 1-band image with palette into RGB image for JPEG output."""
    band = data_array[0, :, :]
    levels = list(palette.pal.keys())
    colors = [
        palette.color_to_color_entry(value, with_alpha=True)
        for value in palette.pal.values()
    ]
    norm = matplotlib.colors.BoundaryNorm(levels, len(levels) - 1)

    # handle palette no data value
    nodata_color = (0, 0, 0, 0)
    if palette.ndv is not None:
        nodata_color = palette.color_to_color_entry(palette.ndv, with_alpha=True)

    # Apply normalization to get palette indices
    indexed_band = norm(band)

    # Create RGB output array
    height, width = band.shape
    rgb_array = np.zeros((3, height, width), dtype='uint8')

    # Apply colors based on palette indices
    for i, color in enumerate(colors):
        mask = indexed_band == i
        rgb_array[0, mask] = color[0]  # Red
        rgb_array[1, mask] = color[1]  # Green
        rgb_array[2, mask] = color[2]  # Blue

    # Handle NaN/nodata values
    nan_mask = np.isnan(band)
    if nan_mask.any():
        rgb_array[0, nan_mask] = nodata_color[0]
        rgb_array[1, nan_mask] = nodata_color[1]
        rgb_array[2, nan_mask] = nodata_color[2]

    return rgb_array, None


def scale_paletted_1band(
    data_array: DataArray, palette: ColorPalette
) -> tuple[ndarray, ColorMap]:
    """Scale a 1-band image with palette into modified image and associated color_map.

    Use the palette's levels and values, transform the input data_array into
    the correct levels indexed from 0-255 return the scaled array along side of
    a colormap corresponding to the new levels.
    """
    band = data_array[0, :, :]
    levels = list(palette.pal.keys())
    colors = [
        palette.color_to_color_entry(value, with_alpha=True)
        for value in palette.pal.values()
    ]
    norm = matplotlib.colors.BoundaryNorm(levels, len(levels) - 1)

    # handle palette no data value
    nodata_color = (0, 0, 0, 0)
    if palette.ndv is not None:
        nodata_color = palette.color_to_color_entry(palette.ndv, with_alpha=True)

    colors = [*colors, nodata_color]

    scaled_band = norm(band)

    # Set underflow and nan values to nodata
    scaled_band[scaled_band == -1] = len(colors) - 1
    scaled_band[np.isnan(band)] = len(colors) - 1

    color_map = colormap_from_colors(colors)
    raster_data = np.expand_dims(scaled_band.data, 0)
    return np.array(raster_data, dtype='uint8'), color_map


def image_driver(mime: str) -> str:
    """Return requested rasterio driver for output image."""
    if re.search('jpeg', mime, re.I):
        return 'JPEG'
    return 'PNG'


def get_color_map_from_image(image: Image) -> dict:
    """Get a writable color map

    Read the RGBA palette from a PIL Image and covert into a dictionary
    that can be written by rasterio.

    """
    color_tuples = np.array(image.getpalette(rawmode='RGBA')).reshape(-1, 4)
    color_map = all_black_color_map()
    for idx, color_tuple in enumerate(color_tuples):
        color_map[idx] = tuple(color_tuple)
    return color_map


def get_aux_xml_filename(image_filename: Path) -> Path:
    """Get aux.xml filenames."""
    return image_filename.with_suffix(image_filename.suffix + '.aux.xml')


def get_tiled_filename(input_file: Path, locator: dict | None = None) -> Path:
    """Add a column, row identifier to the output files.

    Only update if there is a valid locator dict.
    """
    if locator is not None:
        return input_file.with_suffix(
            f'.r{int(locator["row"]):02d}c{int(locator["col"]):02d}{input_file.suffix}'
        )
    return input_file


def output_image_file(input_file_path: Path, driver: str = 'PNG'):
    """Generate the output image name."""
    if driver == 'PNG':
        ext = '.png'
    else:
        ext = '.jpg'
    return input_file_path.with_suffix(ext)


def output_world_file(input_file_path: Path, driver: str = 'PNG'):
    """Generate an output world file name."""
    if driver == 'PNG':
        ext = '.pgw'
    else:
        ext = '.jgw'
    return input_file_path.with_suffix(ext)


def validate_file_crs(data_array: DataArray) -> None:
    """Explicit check for a CRS on the input geotiff.

    Raises HyBIGError if crs is missing.
    """
    if data_array.rio.crs is None:
        raise HyBIGError('Input geotiff must have defined CRS.')


def validate_file_type(dsr: DatasetReader) -> None:
    """Ensure we can work with the input data file.

    Raise an exception if this file is unusable by the service.

    """
    if dsr.driver != 'GTiff':
        raise HyBIGError(f'Input file type not supported: {dsr.driver}')


def get_destination(grid_parameters: GridParams, n_bands: int) -> ndarray:
    """Initialize an array for writing an output raster."""
    return np.zeros(
        (n_bands, grid_parameters['height'], grid_parameters['width']), dtype='uint8'
    )


def write_georaster_as_browse(
    data_array: DataArray,
    raster: ndarray,
    color_map: dict | None,
    grid_parameters: GridParams,
    driver='PNG',
    out_file_name='outfile.png',
    out_world_name='outfile.pgw',
    logger=Logger,
) -> None:
    """Write raster data to output file.

    Writes the raster to an output file using metadata from the original
    source, and over-riding some values based on the input message and derived
    grid_parameters, doing a nearest neighbor resampling to the target grid.

    """
    n_bands = raster.shape[0]

    if color_map is not None:
        dst_nodata = NODATA_IDX
        color_map[dst_nodata] = NODATA_RGBA
    else:
        # for banded data set the each band's destination nodata to zero (TRANSPARENT).
        dst_nodata = TRANSPARENT

    creation_options = {
        **grid_parameters,
        'driver': driver,
        'dtype': 'uint8',
        'count': n_bands,
    }

    dest_array = get_destination(grid_parameters, n_bands)

    logger.info(f'Create output image with options: {creation_options}')

    with rasterio.open(out_file_name, 'w', **creation_options) as dst_raster:
        for dim in range(0, n_bands):
            reproject(
                source=raster[dim, :, :],
                destination=dest_array[dim, :, :],
                src_transform=data_array.rio.transform(),
                src_crs=data_array.rio.crs,
                dst_transform=grid_parameters['transform'],
                dst_crs=grid_parameters['crs'],
                dst_nodata=dst_nodata,
                resampling=Resampling.nearest,
            )

        dst_raster.write(dest_array)
        if color_map is not None:
            dst_raster.write_colormap(1, color_map)

    with open(out_world_name, 'w', encoding='UTF-8') as out_wd:
        out_wd.write(dumpsw(creation_options['transform']))



"""Module containing utility functionality for browse generation."""

from harmony_service_lib.message import Message as HarmonyMessage


def get_harmony_message_from_params(params: dict | None) -> HarmonyMessage:
    """Constructs a harmony message from the input parms.

    We have to create a harmony message to pass to the create_browse_imagery
    function so that both the library and service calls are identical.

    """
    if params is None:
        params = {}
    mime = params.get('mime', 'image/png')
    crs = params.get('crs', None)
    scale_extent = params.get('scale_extent', None)
    scale_size = params.get('scale_size', None)
    height = params.get('height', None)
    width = params.get('width', None)

    return HarmonyMessage(
        {
            'format': {
                'mime': mime,
                'crs': crs,
                'srs': crs,
                'scaleExtent': scale_extent,
                'scaleSize': scale_size,
                'height': height,
                'width': width,
            },
        }
    )



"""Utility routines to aid in determining target size parameters.

This module handles determiniation of the output image size parameters.  These
are geared towards satisfying GIBS requirements, but allows a user to override
most options from the input Harmony Message.

There's a number of "rules" from the GIBS ICD that are codified in this module.

"""

from collections import namedtuple
from typing import TypedDict

import numpy as np
from affine import Affine
from harmony_service_lib.message import Message
from harmony_service_lib.message_utility import (
    has_dimensions,
    has_scale_extents,
    has_scale_sizes,
)

# pylint: disable-next=no-name-in-module
from rasterio.crs import CRS
from rasterio.transform import AffineTransformer, from_bounds, from_origin
from xarray import DataArray

from hybig.crs import (
    choose_target_crs,
)


class GridParams(TypedDict):
    """Convenience to describe a grid parameters dictionary."""

    height: int
    width: int
    crs: CRS
    transform: Affine


class ScaleExtent(TypedDict):
    """Convenience to describe a scale extent dictionary."""

    xmin: float
    ymin: float
    xmax: float
    ymax: float


class ScaleSize(TypedDict):
    """Convenience to describe a scale size dictionary."""

    x: float
    y: float


class Dimensions(TypedDict):
    """Convenience to describe a scale a dimension dictionary."""

    width: int
    height: int


ResolutionInfo = namedtuple(
    'ResolutionInfo',
    [
        'resolution',
        'pixel_size',
        'height',
        'width',
        'overview_levels',
        'overview_scale',
    ],
)

# GIBS output resolution information from 423-ICD-009, Revision A Effecitve
# Date: November 2016

# This is Table 4.1.8-1 WGS84 (EPSG: 4326 Resolutions)
epsg_4326_resolutions = [
    ResolutionInfo('2km', 0.017578125, 10240, 20480, 6, 2),
    ResolutionInfo('1km', 0.0087890625, 20480, 40960, 7, 2),
    ResolutionInfo('500m', 0.00439453125, 40960, 81920, 8, 2),
    ResolutionInfo('250m', 0.002197265625, 81920, 163840, 9, 2),
    ResolutionInfo('125m', 0.0010986328125, 163840, 327680, 10, 2),
    ResolutionInfo('62.5m', 0.00054931640625, 327680, 655360, 11, 2),
    ResolutionInfo('31.25m', 0.000274658203125, 655360, 1310720, 12, 2),
    ResolutionInfo('15.625m', 0.0001373291015625, 1310720, 2521440, 13, 2),
]

# Conversion used above: resolution / pixel_size
METERS_PER_DEGREE = 113777.77777777778

# This is Table 4.1.8-2 NSIDC Sea Ice Polar Stereographic Extent (EPSG:3413) Resolutions
epsg_3413_resolutions = [
    ResolutionInfo('2km', 2048, 4096, 4096, 3, 2),
    ResolutionInfo('1km', 1024, 8192, 8192, 4, 2),
    ResolutionInfo('500m', 512, 16384, 16384, 5, 2),
    ResolutionInfo('250m', 256, 32768, 32768, 6, 2),
    ResolutionInfo('125m', 128, 65536, 65536, 7, 2),
    ResolutionInfo('62.5m', 64, 131072, 131072, 8, 2),
    ResolutionInfo('31.25m', 32, 252144, 252144, 9, 2),
    ResolutionInfo('15.625m', 16, 524288, 524288, 10, 2),
]

# The Antarctic resolutions match the northern resolutions precisely.
# Table 4.1.8-3. Antarctic Polar Stereographic (EPSG: 3031) Resolutions
epsg_3031_resolutions = epsg_3413_resolutions


def get_target_grid_parameters(message: Message, data_array: DataArray) -> GridParams:
    """Get the output image parameters.

    This computes the target grid of the ouptut image. The grid is defined by
    of the extent of the output: ScaleExtents, and either the grid cell sizes:
    ScaleSizes or the dimensions: width and height.

    - User submitted parameters take precedence over computed
    parameters.

    - Computed parameters attempt to generate GIBS suitable images.

    """
    target_crs = choose_target_crs(message.format.srs, data_array)
    target_scale_extent = choose_scale_extent(message, target_crs, data_array)
    target_dimensions = choose_target_dimensions(
        message, data_array, target_scale_extent, target_crs
    )
    return get_rasterio_parameters(target_crs, target_scale_extent, target_dimensions)


def choose_scale_extent(
    message: Message, target_crs: CRS, data_array: DataArray
) -> ScaleExtent:
    """Return the scaleExtent for the target image.

    Returns a scale extent found in the input Message.

    Otherwise, computes a bounding box in the target CRS based on the input
    granule extent.

    """
    if has_scale_extents(message):
        # These values must be in the target_crs projection.
        scale_extent = ScaleExtent(
            {
                'xmin': message.format.scaleExtent.x.min,
                'ymin': message.format.scaleExtent.y.min,
                'xmax': message.format.scaleExtent.x.max,
                'ymax': message.format.scaleExtent.y.max,
            }
        )
    else:
        left, bottom, right, top = data_array.rio.transform_bounds(target_crs)

        # Correct for antimeridian crossing.
        if left > right:
            right = right + 360
        scale_extent = ScaleExtent(
            {'xmin': left, 'ymin': bottom, 'xmax': right, 'ymax': top}
        )
    return scale_extent


def choose_target_dimensions(
    message: Message, data_array: DataArray, scale_extent: ScaleExtent, target_crs: CRS
) -> Dimensions:
    """This selects or computes the target Dimensions.

    This routine finalizes the output grid.  The target dimensions are
    returned.  These along with the CRS and scaleExent will be used to
    selecting the correct output transformation for the final image.

    To determine the target dimensions the following occurs.

    The input harmony message format is searched for dimensions or scaleSizes:
    If dimensions (height and width) are found, they are returned.
    If scaleSize is found, it is converted to height and width and returned.

    If no scaleSize or dimensions are input. The input dataset metadata
    resolution is checked against the GIBS preferred resolutions and the best
    fit resolution along with the scaleExtent is used to compute the target
    dimensions to return.

    """
    if has_dimensions(message):
        dimensions = Dimensions(
            {'height': message.format.height, 'width': message.format.width}
        )
    elif has_scale_sizes(message):
        dimensions = compute_target_dimensions(
            scale_extent, message.format.scaleSize.x, message.format.scaleSize.y
        )
    else:
        dimensions = best_guess_target_dimensions(data_array, scale_extent, target_crs)

    return dimensions


def get_rasterio_parameters(
    crs: CRS, scale_extent: ScaleExtent, dimensions: Dimensions
) -> GridParams:
    """Convert the grid into rasterio consumable format.

    Returns a Dictionary of keyword params suitable for rasterio to use in
    writing an output image.

    To write images, rasterio wants parameters: 'width', 'height', 'crs', and
    'transform'

    """
    transform = from_bounds(
        scale_extent['xmin'],
        scale_extent['ymin'],
        scale_extent['xmax'],
        scale_extent['ymax'],
        dimensions['width'],
        dimensions['height'],
    )
    return {
        'width': dimensions['width'],
        'height': dimensions['height'],
        'crs': crs,
        'transform': transform,
    }


def create_tiled_output_parameters(
    grid_parameters: GridParams,
) -> tuple[list[GridParams], list[dict] | list[None]]:
    """Split the output grid if necessary.

    When the number of grid cells exceeds 8192x8192, we tile the output
    to a reasonable size: 4096x4096

    returns a list of GridParams object that completely cover the input scale
    extent tiled as needed.

    """
    if not needs_tiling(grid_parameters):
        return [grid_parameters], [None]

    crs = grid_parameters['crs']
    full_width = grid_parameters['width']
    full_height = grid_parameters['height']
    transform = grid_parameters['transform']
    resolution_x = transform.a
    resolution_y = np.abs(transform.e)

    cells_per_tile_width = get_cells_per_tile()
    cells_per_tile_height = get_cells_per_tile()

    width_origins = compute_tile_boundaries(cells_per_tile_width, full_width)
    height_origins = compute_tile_boundaries(cells_per_tile_height, full_height)
    width_dimensions = compute_tile_dimensions(width_origins)
    height_dimensions = compute_tile_dimensions(height_origins)

    transformer = AffineTransformer(transform)

    grid_parameter_list = []
    tile_locator = []
    for h_idx, row in enumerate(height_origins):
        for w_idx, col in enumerate(width_origins):
            height = height_dimensions[h_idx]
            width = width_dimensions[w_idx]
            x_loc, y_loc = transformer.xy(row, col, offset='ul')
            tile_grid_transform = from_origin(x_loc, y_loc, resolution_x, resolution_y)
            if height > 0 and width > 0:
                grid_parameter_list.append(
                    GridParams(
                        {
                            'width': width,
                            'height': height,
                            'crs': crs,
                            'transform': tile_grid_transform,
                        }
                    )
                )
                tile_locator.append({'row': h_idx, 'col': w_idx})

    return grid_parameter_list, tile_locator


def compute_tile_dimensions(origins: list[int]) -> list[int]:
    """Return a list of tile dimensions.

    From a list of origin locations, return the dimension for each tile.

    """
    return list(np.diff(origins, append=origins[-1]).astype('int'))


def compute_tile_boundaries(target_size: int, full_size: int) -> list[int]:
    """Returns a list of boundary cells.

    The returned boundary cells are the column [or row] values for each of the
    output tiles. They should always start at 0, and end at the full_size
    input, stepping by target size until the last step, which can be any size
    including 0.

    """
    n_boundaries = int(full_size // target_size) + 1
    boundaries = [target_size * index for index in range(n_boundaries)]

    if boundaries[-1] != full_size:
        boundaries.append(full_size)

    return boundaries


def get_cells_per_tile() -> int:
    """Optimum cells per tile.

    From discussions this is chosen to be 4096, so that any image that is tiled
    will end up with 4096x4096 gridcell tiles.
    """
    return 4096


def needs_tiling(grid_parameters: GridParams) -> bool:
    """Returns true if the grid is too large for GIBS.

    From discussion, this limit is set to 8192*8192 cells.
    """
    MAX_UNTILED_GRIDCELLS = 8192 * 8192
    return grid_parameters['height'] * grid_parameters['width'] > MAX_UNTILED_GRIDCELLS


def best_guess_target_dimensions(
    data_array: DataArray, scale_extent: ScaleExtent, target_crs: CRS
) -> Dimensions:
    """Return best guess for output image dimensions.

    Using the information from the scaleExtent and the input dataset metadata,
    compute the height and width appropriate for the output image.

    North and South projections have matching resolutions so we just use the
    3413 resolutions for all projected data.
    """
    resolution_list = None
    if not target_crs.is_projected:
        resolution_list = epsg_4326_resolutions
    else:
        resolution_list = epsg_3413_resolutions

    x_res, y_res = resolution_in_target_crs_units(data_array, target_crs)
    return guess_dimensions(x_res, y_res, scale_extent, resolution_list)


def resolution_in_target_crs_units(
    data_array: DataArray, target_crs: CRS
) -> tuple[float, float]:
    """Return the x and y target resolutions

    The input resolution can be determined from the Affine transformation, but
    if the input dataset is projected, and the target CRS is unprojected, we
    need to convert the resolution meters into degrees.

    This routine is only called for best_guess_target_dimensions which means
    the user has not supplied any input parameters and we are trying to
    determine the dimensions for the output image.
    """
    if data_array.rio.crs.is_projected == target_crs.is_projected:
        x_res = data_array.rio.transform().a
        y_res = abs(data_array.rio.transform().e)
    elif target_crs.is_projected:
        # transform from latlon to meters
        x_res = data_array.rio.transform().a * METERS_PER_DEGREE
        y_res = abs(data_array.rio.transform().e) * METERS_PER_DEGREE
    else:
        # transform from meters to lat/lon
        x_res = data_array.rio.transform().a / METERS_PER_DEGREE
        y_res = abs(data_array.rio.transform().e) / METERS_PER_DEGREE

    return x_res, y_res


def guess_dimensions(
    x_res: float,
    y_res: float,
    scale_extent: ScaleExtent,
    icd_resolution_list: list[ResolutionInfo],
) -> Dimensions:
    """Guess Dimensions given the input information and ICD resolutions."""
    coarsest_resolution = icd_resolution_list[0].pixel_size

    if x_res > coarsest_resolution and y_res > coarsest_resolution:
        dimensions = compute_target_dimensions(scale_extent, x_res, y_res)
    else:
        target_res = find_closest_resolution(list({x_res, y_res}), icd_resolution_list)
        dimensions = compute_target_dimensions(
            scale_extent, target_res.pixel_size, target_res.pixel_size
        )
    return dimensions


def compute_target_dimensions(
    scale_extent: ScaleExtent, x_res: float, y_res: float
) -> Dimensions:
    """Compute dimensions from inputs.

    We round the computed dimensions because we need to have integer values of
    height and width. Because of the rounding, this function could return a
    height/width that in combination with the ScaleExtent would be slightly
    inconsistent with the input res (ScaleSize).  This is ok, because for
    computed values of scaleExtent, we are already self consistent.  If there's
    a discrepancy, the user has described a scaleExtent in the input
    HarmonyMessage and has *not* specified a scaleSize. So we honor the
    scaleExtent in a previous function.

    """
    return {
        'height': round((scale_extent['ymax'] - scale_extent['ymin']) / y_res),
        'width': round((scale_extent['xmax'] - scale_extent['xmin']) / x_res),
    }


def find_closest_resolution(
    resolutions: list[float], resolution_info: list[ResolutionInfo]
) -> ResolutionInfo | None:
    """Return closest match to GIBS preferred Resolution Info.

    Cycle through all input resolutions and return the resolution_info that has
    the smallest absolute difference to any of the input resolutions.

    """
    best_info = None
    smallest_diff = np.inf
    for res in resolutions:
        for info in resolution_info:
            resolution_diff = np.abs(res - info.pixel_size)
            if resolution_diff < smallest_diff:
                smallest_diff = resolution_diff
                best_info = info

    return best_info



"""Module defining custom exceptions."""


class HyBIGError(Exception):
    """Base error class for exceptions raised by HyBIG library."""

    def __init__(self, message=None):
        """All HyBIG errors have a message field."""
        self.message = message


class HyBIGNoColorInformation(HyBIGError):
    """Used to describe missing color information."""


class HyBIGValueError(HyBIGError):
    """Input was incorrect for the routine."""
